// the texture
TEXTURE *flub_text;
long text_offs;

// small buffer of 256x1 containing the light pattern (fake phong ;)
unsigned char *light;

// our 16 bit zbuffer, we don't need it much in this example, unless
// the flubber has a really weird shape
unsigned short *zbuffer;

// the look-up table
unsigned char *flub_lut;

// some constants, number of spans per slice
#define SPANS 32
#define SPANMASK 31

// store points as new data type, the type vector would take up more
// space than necessary
    struct {
      float x, z;
    } fpts[200][SPANS], fnrm[200][SPANS];



/*
 * draw a horizontal double textured span
 */
void DrawSpan( int y, int x1, int tx1, int px1, int z1,
                      int x2, int tx2, int px2, int z2 )
{
  // quick check, if facing back then return
     if (x1>=x2) return;
  // compute deltas for interpolation
     int dx = x2-x1,
         dtx = (tx2-tx1)/dx,  // assume 16.16 fixed point
         dpx = (px2-px1)/dx,  // 16.16
         dz  = (z2-z1)/dx;    // doesn't matter, whatever parameters are is fine
  // get destination offset in buffer
     long offs = y*320+x1;
  // loop for all pixels concerned
     for (int i=0; i<dx; i++)
     {
      // check z buffer
         if (z1<zbuffer[offs])
         {
         // if visible load the texel from the translated texture
            int texel = flub_text->location[((y+text_offs<<8)&0xff00)+((tx1>>16)&0xff)],
         // and the texel from the light map
                lumel = light[(px1>>16)&0xff];
         // mix them together
            vga->page_draw[offs] = flub_lut[(texel<<8)+lumel];
         // and update the zbuffer
            zbuffer[offs] = z1;
         }
      // interpolate our values
         px1 += dpx;
         tx1 += dtx;
         z1  += dz;
      // and find next pixel
         offs++;
     }
}

/*
 * naively scan all the colours to find the closest one
 */
unsigned char Closest_Colour( unsigned char *pal, unsigned char r,
                              unsigned char g, unsigned char b )
{
     long dist = 1<<30, newDist;
     unsigned char index;
  // for all colours loop
     for (int i=0; i<256; i++)
     {
      // calculate how far the current colour is from the required colour
         newDist = (r-pal[i*3])*(r-pal[i*3])
                 + (g-pal[i*3+1])*(g-pal[i*3+1])
                 + (b-pal[i*3+2])*(b-pal[i*3+2]);
      // if it's the same then we've found a match, so return it
         if (newDist==0) return i;
      // if the distance is closer, then remember this colour's index
         if (newDist<dist) {
            index = i;
            dist = newDist;
         }
     }
  // return the one we found closest
     return index;
}

/*
 * given a palette compute the best lookup table containing 256 shades
 * of each original colour
 */
void Compute_LUT1( unsigned char *pal )
{
   unsigned char r, g, b, l;
   int i;
// for each colour
   for (int j=0; j<256; j++)
   {
       r = pal[j*3];
       g = pal[j*3+1];
       b = pal[j*3+2];
    // calculate shades from a third of the colour to the colour
       for (i=0; i<240; i++)
       {
        // get the index of the best colour
           flub_lut[(j<<8)+i] = Closest_Colour( pal, r*(120+i)/360, g*(120+i)/360, b*(120+i)/360 );
       }
    // calc shades half way from the colour to white
       for (i=0; i<16; i++)
       {
           flub_lut[(j<<8)+240+i] = Closest_Colour( pal, r+(255-r)*i/31, g+(255-g)*i/31, b+(255-b)*i/31 );
       }

   }
}


void Flubber_Init()
{
    int i, j;
 // load the texture
    flub_text = new TEXTURE( "flubber.png" );
 // prepare the lighting
    light = new unsigned char[256];
    for (i=0; i<256; i++)
    {
     // calculate distance from the centre
        int c = abs(255-i*2);
     // check for overflow
        if (c>255) c = 255;
     // store lumel
        light[i] = 255-c;
    }
// compute the lookup table
    flub_lut = new (unsigned char)[65536];
    Compute_LUT1( flub_text->palette );
// prepare 3D data
    zbuffer = new unsigned short[64000];
// initialise each point of each slice
    for (i=0; i<SPANS; i++)
    {
        float angle = (float)i*M_PI*2/SPANS,
              ca = cos( angle ),
              sa = sin( angle );
        for (j=0; j<200; j++)
        {
// use these equations to get a weird shaped flubber
//        float radx = 48 + 32 * sin( ((float)i+47*cos((float)j*0.123)) * 0.019 )
//                        * cos( ((float)j+39*sin((float)i*0.137)) * 0.023 ),
//              rady = 48 + 32 * cos( ((float)i+31*sin((float)j*0.147)) * 0.027 )
//                         * sin( ((float)j+43*cos((float)i*0.111)) * 0.014 );
         // this generates a flat cylinder
            float radx = 32, rady = 96;
         // store the points
            fpts[j][i].x = radx * ca;
            fpts[j][i].z = rady * sa;
        }
    }
 // now calculate the normals
    for (i=0; i<SPANS; i++)
    {
        for (j=0; j<200; j++)
        {
         // get coords of tangeant vector
            float nx = fpts[j][(i+1)&SPANMASK].x - fpts[j][(i-1)&SPANMASK].x,
                  nz = fpts[j][(i+1)&SPANMASK].z - fpts[j][(i-1)&SPANMASK].z,
         // and compute it's length
                  inorm = 1/sqrt(nx*nx+nz*nz);
         // now make sure the vector has length one
            nx *= inorm;
            nz *= inorm;
         // and store it
            fnrm[j][i].x = nx;
            fnrm[j][i].z = nz;
        }
    }
}

void Flubber_Run()
{
    int i, j;
    bQuit = false;
// run the main loop
    while (!kbhit() && !bQuit)
    {
        // set the colours
          for (i=0; i<256; i++)
          {
              vga->SetColour( i,
              (unsigned char)((flub_text->palette[i*3]>>2)*(1-flash_coef)+63*flash_coef),
              (unsigned char)((flub_text->palette[i*3+1]>>2)*(1-flash_coef)+63*flash_coef),
              (unsigned char)((flub_text->palette[i*3+2]>>2)*(1-flash_coef)+63*flash_coef) );
          }
          long long currentTime = timer->getCount();
       // clear the zbuffer
          memset( zbuffer, 255, 128000 );
       // and determine how far to scroll the texture
          text_offs = currentTime / 31456;
       // clear the background
          vga->Clear();
       // draw all slices
          for (i=0; i<200; i++)
          {
              int xc[SPANS], zc[SPANS], nc[SPANS];
           // set a different rotation angle for each slice
              float angle = (float)currentTime / 1948613 +
              2 * M_PI * (cos( (float)currentTime/3179640 )
              * cos( -(float)i / 193 ) * sin( (float)currentTime/2714147 ) + 1 );
           // compute the sine and cosine in 8.8 fxd point
              float ca = 256 * cos( angle ),
                    sa = 256 * sin( angle );
           // and determine the horizontal position of this slice
              long xoffs = (long)(40960+10240 * cos( (float)currentTime/2945174 )
                                        * sin( (float)i/59 ) );
              for (j=0; j<SPANS; j++)
              {
               // rotate the point and get X coordinate
                  xc[j] = xoffs+(long)( fpts[i][j].x * ca + fpts[i][j].z * -sa );
               // rotate the point and get Z coordinate
                  zc[j] = (long)( fpts[i][j].x * sa + fpts[i][j].z * ca );
               // now get the coordinate in the lightmap by computing
               // the X component of the normal
                  long tmp = (long)( fnrm[i][j].x * ca + fnrm[i][j].z * -sa );
                  tmp = 256-tmp;
                  nc[j] = (128<<16)+(tmp<<15);
              }
           // now just draw all spans
              for (j=0; j<SPANS; j++)
              {
                  DrawSpan( i, xc[j]>>8, j<<20, nc[j], 49152+zc[j],
                            xc[(j+1)&SPANMASK]>>8, (j+1)<<20,
                            nc[(j+1)&SPANMASK], 49152+zc[(j+1)&SPANMASK] );
              }
          }
       // flush to video ram
          vga->Update();
          frameCount++;
    }
}

void Flubber_Done()
{
    delete flub_text;
    delete zbuffer;
    delete flub_lut;
}

